from typing import List


class EmbedMenuEmojiConfig:
    def __init__(self, delete_message: str = "\N{CROSS MARK}",
                 unsupported_transition: str = "\N{NO ENTRY SIGN}"):
        self.delete_message = delete_message
        self.unsupported_transition = unsupported_transition

    def to_list(self) -> List[str]:
        return [self.delete_message, self.unsupported_transition]


DEFAULT_EMBED_MENU_EMOJI_CONFIG = EmbedMenuEmojiConfig()



from typing import Iterable, Optional, List, Union

from discord import Reaction, Message, Member, RawReactionActionEvent, Emoji

from discordmenu.embed.emoji import DEFAULT_EMBED_MENU_EMOJI_CONFIG, EmbedMenuEmojiConfig
from discordmenu.emoji.emoji import discord_emoji_to_emoji_name


class ReactionFilter:
    def __init__(self, reaction_filter: Optional["ReactionFilter"] = None):
        self.inner_filter = reaction_filter

    async def allow_reaction(self, message: Message, reaction: Reaction, member: Member) -> bool:
        parent = await self._allow_reaction(message, reaction, member)
        if parent:
            return True

        return await self.inner_filter.allow_reaction(message, reaction, member) if self.inner_filter else False

    async def allow_reaction_raw(self, message: Message, reaction: RawReactionActionEvent) -> bool:
        parent = await self._allow_reaction_raw(message, reaction)
        if parent:
            return True

        return await self.inner_filter.allow_reaction_raw(message, reaction) if self.inner_filter else False

    async def _allow_reaction(self, message: Message, reaction: Reaction, member: Member) -> bool:
        return True

    async def _allow_reaction_raw(self, message: Message, reaction: RawReactionActionEvent) -> bool:
        return True


class ValidEmojiReactionFilter(ReactionFilter):
    def __init__(self, valid_emoji_names: Iterable[Union[str, Emoji]],
                 default_emoji_override: Optional[EmbedMenuEmojiConfig] = None,
                 filters: Optional[ReactionFilter] = None):
        super().__init__(filters)
        default_emojis = default_emoji_override or DEFAULT_EMBED_MENU_EMOJI_CONFIG
        emoji_set = set(default_emojis.to_list())
        emoji_set.update(valid_emoji_names)
        self.valid_emoji_names = list(emoji_set)

    async def _allow_reaction(self, message: Message, reaction: Reaction, member: Member) -> bool:
        valid_emoji_reaction = discord_emoji_to_emoji_name(reaction.emoji) in self.valid_emoji_names
        return valid_emoji_reaction

    async def _allow_reaction_raw(self, message: Message, reaction: RawReactionActionEvent) -> bool:
        valid_emoji_reaction = discord_emoji_to_emoji_name(reaction.emoji) in self.valid_emoji_names
        return valid_emoji_reaction


class BotAuthoredMessageReactionFilter(ReactionFilter):
    """
    This prevents the bot from reacting to messages it didn't post.
    """

    def __init__(self, bot_id: int, filters: Optional[ReactionFilter] = None):
        super().__init__(filters)
        self.bot_id = bot_id

    async def _allow_reaction(self, message: Message, reaction: Reaction, member: Member):
        return message.author.id == self.bot_id

    async def _allow_reaction_raw(self, message: Message, reaction: RawReactionActionEvent):
        return message.author.id == self.bot_id


class NotPosterEmojiReactionFilter(ReactionFilter):
    """
    This prevents the bot from reacting to its own emojis in DM.
    """

    async def _allow_reaction(self, message: Message, reaction: Reaction, member: Member):
        return message.author.id != member.id

    async def _allow_reaction_raw(self, message: Message, reaction: RawReactionActionEvent) -> bool:
        if not reaction.guild_id:
            return message.author.id != reaction.user_id
        return True


class MessageOwnerReactionFilter(ReactionFilter):
    def __init__(self, original_author_id: int, filters: Optional[ReactionFilter] = None):
        super().__init__(filters)
        self.original_author_id = original_author_id

    async def _allow_reaction(self, message: Message, reaction: Reaction, member: Member) -> bool:
        return member.id == self.original_author_id

    async def _allow_reaction_raw(self, message: Message, reaction: RawReactionActionEvent) -> bool:
        if reaction.guild_id:
            return reaction.member.id == self.original_author_id
        return True


class FriendReactionFilter(ReactionFilter):
    def __init__(self, original_author_id: int, friends_ids: List[int], filters: Optional[ReactionFilter] = None):
        super().__init__(filters)
        self.original_author_id = original_author_id
        self.friend_ids = friends_ids

    async def _allow_reaction(self, message: Message, reaction: Reaction, member: Member) -> bool:
        # return await friend_ids(self.original_author_id, member.id)
        return member.id in self.friend_ids

    async def _allow_reaction_raw(self, message: Message, reaction: RawReactionActionEvent) -> bool:
        # return self.friend_ids(self.original_author_id, reaction.member.id)
        return reaction.member.id in self.friend_ids



from typing import List, Union, Sequence

from discord import Emoji as Emoji
from discord.ext.commands import Bot


def is_ascii(s):
    return all(ord(c) < 128 for c in s)


class EmojiCache:
    def __init__(self, guild_ids: List[int]):
        self.guild_ids = guild_ids
        self.custom_emojis: List[Emoji] = []

    def set_guild_ids(self, guild_ids: List[int]) -> None:
        self.guild_ids = guild_ids

    def refresh_from_discord_bot(self, bot: Bot) -> None:
        self.custom_emojis = [e for g in bot.guilds if g.id in self.guild_ids for e in g.emojis]

    def refresh_from_emojis(self, emojis: List[Emoji]) -> None:
        self.custom_emojis = emojis

    def get_by_name(self, names: Union[str, Sequence[str]]) -> Union[Emoji, str]:
        return self._get_value_by_name(names, self._emoji)

    def get_name_by_name(self, names: Union[str, Sequence[str]]) -> Union[Emoji, str]:
        return self._get_value_by_name(names, self._emoji_name)

    def _get_value_by_name(self, names: Union[str, Sequence[str]], f) -> Union[Emoji, str]:
        # names will be a tuple if looking up immediately from an emoji list
        # but it becomes a list when deserialized from an ims
        if isinstance(names, str):
            res = [f(e) for e in self.custom_emojis if e.name == names]
            return res[0] if len(res) > 0 else names
        else:
            # return the first success, if there is one
            # otherwise just return the last thing, which is presumed to be a default emoji
            # a list of length 0 is invalid input
            if len(names) == 0:
                raise KeyError
            for name in names:
                res = [e.name for e in self.custom_emojis if e.name == name]
                if len(res) > 0:
                    return res[0]
            return self.get_by_name(names[-1])

    @staticmethod
    def _emoji(e: Emoji) -> Emoji:
        return e

    @staticmethod
    def _emoji_name(e: Emoji) -> str:
        return e.name

    def get_emoji(self, name: str) -> str:
        """Use for getting an emoji name to print in text"""
        for e in self.custom_emojis:
            if e.name == name:
                return str(e)

        # special case when a unicode character is specified
        if len(name) == 1 and not is_ascii(name):
            return name

        return ":{}:".format(name)

    def get_raw_emoji(self, name: str) -> str:
        """Same as get_emoji but no colons - use for getting a reaction name"""
        for e in self.custom_emojis:
            if e.name == name:
                return str(e)

        # special case when a unicode character is specified
        if len(name) == 1 and not is_ascii(name):
            return name

        return str(name)


emoji_cache = EmojiCache([])



from typing import List, Union, Sequence, Iterable

from discord import Message

from discordmenu.embed.view import EmbedView
from discordmenu.emoji.emoji_cache import emoji_cache


class EmbedControl:
    def __init__(self, embed_views: List[EmbedView],
                 emoji_buttons: Iterable[Union[str, Sequence[str]]] = None):
        self.embed_views = embed_views
        self.emoji_buttons = [emoji_cache.get_name_by_name(e) for e in emoji_buttons or []]

    @staticmethod
    def from_message(message: Message) -> "EmbedControl":
        emojis = [r.emoji for r in message.reactions]
        embed_views = [EmbedView.from_message(e) for e in message.embeds]
        return EmbedControl(embed_views, emojis)



import asyncio
from typing import Dict, List, Optional, Union, Sequence

from discord import Embed, Emoji, Forbidden, Message
from discord import NotFound
from discord.ext.commands import Context

from discordmenu.embed.control import EmbedControl
from discordmenu.embed.view import EmbedView
from discordmenu.emoji.emoji_cache import emoji_cache


async def update_message(message: Message, updated_messaged_contents, guild_message: bool,
                         emoji_diff: Optional[Dict[str, List[Union[str, Emoji]]]] = None) -> None:
    if isinstance(updated_messaged_contents, Embed):
        await message.edit(embed=updated_messaged_contents)
    else:
        await message.edit(content=updated_messaged_contents)
    if emoji_diff:
        add = [message.add_reaction(e) for e in emoji_diff.get('add', [])]
        await asyncio.gather(*add)

        if guild_message:
            remove = [message.clear_reaction(e) for e in emoji_diff.get('remove', [])]
            await asyncio.gather(*remove)


async def remove_reaction(message: Message, emoji: str, user_id: int) -> None:
    if not message.guild:
        # bots don't have permission to delete reactions in DM. So check if theres a guild associated before attempting.
        return

    member = message.guild.get_member(user_id)

    try:
        # support custom emojis
        await message.remove_reaction(emoji_cache.get_raw_emoji(emoji), member)
    except Forbidden:
        pass


async def send_embed_control(ctx: Context, embed_control: EmbedControl, message: Optional[Message] = None) -> Message:
    if type(embed_control.embed_views[0]) != EmbedView:
        raise TypeError("Check return type of your View, an EmbedView is not being returned")
    new_embed = embed_control.embed_views[0].to_embed()
    if message is None:
        message = await ctx.send(embed=new_embed)
    else:
        await message.edit(embed=new_embed)

    emoji_to_add = [emoji_cache.get_by_name(e) for e in embed_control.emoji_buttons]
    add = [message.add_reaction(e) for e in emoji_to_add]
    try:
        await asyncio.gather(*add)
    except NotFound:
        # if messsage is deleted early
        pass
    return message


async def update_embed_control(message: Message, next_embed_control: EmbedControl,
                               emoji_diff: Dict[str, List[Union[Emoji, str]]]) -> None:
    guild_message = bool(message.guild)

    if not next_embed_control:
        await message.delete()

    updated_message_contents = next_embed_control.embed_views[0].to_embed()
    await message.edit(embed=updated_message_contents)

    if emoji_diff:
        add = [message.add_reaction(e) for e in emoji_diff.get('add', [])]
        try:
            await asyncio.gather(*add)
        except NotFound:
            # if messsage is deleted early
            pass

        if guild_message:
            remove = [message.clear_reaction(e) for e in emoji_diff.get('remove', [])]
            try:
                await asyncio.gather(*remove)
            except NotFound:
                # if messsage is deleted early
                pass


def diff_emojis(message: Message, next_embed_control: EmbedControl) -> Dict[str, List[Union[str, Emoji]]]:
    current_emojis = [e.emoji for e in message.reactions]
    next_emojis = next_embed_control.emoji_buttons
    return diff_emojis_raw(current_emojis, next_emojis)


def diff_emojis_raw(current_emojis: List[Union[str, Emoji]], next_emojis: List[Union[str, Emoji]]) \
                -> Dict[str, List[Union[str, Emoji]]]:
    add = sorted(
        set(e for e in next_emojis if e and not emoji_matches(e, current_emojis)),
        key=lambda x: next_emojis.index(x))
    remove = list(set(e for e in current_emojis if e and not emoji_matches(e, next_emojis)))
    return {
        'add': [emoji_cache.get_by_name(e) for e in add],
        'remove': remove,
    }


def emoji_matches(emoji: Union[str, Emoji, Sequence[Union[str, Emoji]]],
                  emoji_to_match: List[Union[str, Emoji]]) -> bool:
    if isinstance(emoji, tuple) or isinstance(emoji, list):
        # handle case of fallback emojis
        return any(emoji_matches(e, emoji_to_match) for e in emoji)
    # handle case of custom emojis as well as normal unicode emojis
    if isinstance(emoji, Emoji):
        emoji = emoji.name
    emoji_to_match = [e.name if isinstance(e, Emoji) else e for e in emoji_to_match]
    return emoji in emoji_to_match
